# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
This folders contains implementation for various interfaces
which will be used by NeST users.
"""

from .base_interface import BaseInterface
from .interface import Interface, create_veth_pair
