# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2023 NITK Surathkal

"""Congestion Algorithm list for TCP/MPTCP Experiments"""

congestion_algo_list = [
    "bbr",
    "bic",
    "cdg",
    "cubic",
    "dctcp",
    "highspeed",
    "htcp",
    "illinois",
    "reno",
    "scalable",
    "vegas",
    "veno",
    "westwood",
    "yeah",
]
